# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Module to access the Fuji Xerox Working Folder server with REST API."""


try:
    import json
except:
    import simplejson as json

import os
import time
import stat
import calendar
import re
import unicodedata
import httplib
import StringIO
from urllib import (urlencode, quote)
from urllib2 import (HTTPError,)
from ionas.restapi.accessinfo import (AccessInfo, RequestInfo)
from ionas.restapi.pathid import (PathIdMap)
from pkgaddon.fxwf_config import (PRIVATE_DRAWER_NAME, HTTP_USER_AGENT_TMPL)


# data_type
NO_DATA = 0
DATA_JSON = 1
DATA_BIN = 2

# fetch_type
FETCH_DRAWER = 0
FETCH_STAT = 1

POST_METHOD = "POST"
PUT_METHOD = "PUT"
DELETE_METHOD = "DELETE"

ROOT_PATH = "/"
OBJECT_URL_TMPL = "/object/%s"
OBJECT_CHILDREN_URL_TMPL = OBJECT_URL_TMPL + "/children"
MAKE_DIRECTORY_URL_TMPL = OBJECT_URL_TMPL + "/new/folder"
MOVE_OBJECT_URL_TMPL = OBJECT_URL_TMPL + "/move"
FILE_READ_URL_TMPL = OBJECT_URL_TMPL + "/content"
FILE_WRITE_URL_TMPL = OBJECT_URL_TMPL + "/new/document"
CABINET_ID = "cabinet"

PRIVATE_DRAWER_ID_PREFIX = "privateDrawer-"
PUBLIC_DRAWER_ID_PREFIX = "publicDrawer-"
FOLDER_ID_PREFIX = "folder-"

ITEMS_NAME = "items"
NAME_NAME = "name"
STAT_NAME = "stat"
OBJECT_ID_NAME = "objectId"
OBJECT_NAME_NAME = "objectName"
DESCRIPTION_NAME = "description"
MODIFIED_DATE_NAME = "modifiedDate"
ORIGINAL_CONTENT_SIZE_NAME = "originalContentSize"

BASE_MODE = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH
FILE_MODE = BASE_MODE | stat.S_IFREG
DIR_MODE = (BASE_MODE | stat.S_IFDIR |
                stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
MTIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

INTERNAL_ERROR_FILE = "intarnal"


def _get_error_stream(err_code, msg, data=None):
    """Return a stream include the JSON format error information."""

    err_data = {"code": err_code, "message": msg}
    if data != None:
        err_data["data"] = data

    err_stream = StringIO.StringIO()
    err_stream.write(json.dumps(err_data))
    err_stream.seek(0)
    return err_stream


class ErrCode:
    """Class for defining the error code."""

    DRAWER_NOT_FOUND = "internalDrawerNotFound"
    OBJECT_NOT_FOUND = "objectNotFound"


class DrawerNotFoundHttpError(HTTPError):
    """Class for the dummy "HTTPError" representing the internal error."""

    def __init__(self):
        """Constructor."""

        HTTPError.__init__(self, INTERNAL_ERROR_FILE,
                            httplib.NOT_FOUND,
                            httplib.responses.get(httplib.NOT_FOUND, ""), {},
                            _get_error_stream(ErrCode.DRAWER_NOT_FOUND, ""))


class ObjectNotFoundHttpError(HTTPError):
    """Class for the dummy "HTTPError" representing the internal error."""

    def __init__(self):
        """Constructor."""

        HTTPError.__init__(self, INTERNAL_ERROR_FILE,
                            httplib.NOT_FOUND,
                            httplib.responses.get(httplib.NOT_FOUND, ""), {},
                            _get_error_stream(ErrCode.OBJECT_NOT_FOUND, ""))


class FxwfInfo(AccessInfo):
    """Class to access to the Fuji Xerox Working Folder server."""

    def __get_session_id(self, cookie_value):
        """Return a session id string in "cookie_value."""

        match = re.search("(JSESSIONID=[^;]+);", cookie_value)
        if match == None:
            return ""
        return match.group(1)

    def __get_headers(self, data_type=NO_DATA, length=0):
        """Return headers used in common requests."""

        headers = {"User-Agent": HTTP_USER_AGENT_TMPL % self.client_id}
        content_type = ""
        if data_type == DATA_JSON:
            content_type = "application/json; charset=UTF-8"
        elif data_type == DATA_BIN:
            content_type = "application/octet-stream; charset=UTF-8"

        if content_type:
            headers["Content-Type"] = content_type
            if length:
                headers["Content-length"] = str(length)

        if self.session_id:
            headers["Cookie"] = self.session_id

        return headers

    def __to_str(self, input):
        """Return string accoding to input."""

        if hasattr(input, "read"):
            return input.read()

        return input

    def __to_utf8(self, string):
        """Convert "string" to UTF-8 string if it "Unicode"."""

        if isinstance(string, unicode):
            return string.encode("UTF-8")
        return string

    def __quote(self, string):
        """Convert "string" to UTF-8 string if it "Unicode" and quote it."""

        return quote(self.__to_utf8(string))

    def __get_fetch_param(self, fetch_type):
        """Return parameter to fetching attributes."""

        FETCH_KEY = "fetch"
        FETCH_ATTRS_DICT = {
                FETCH_DRAWER: (OBJECT_ID_NAME, OBJECT_NAME_NAME,
                                DESCRIPTION_NAME),
                FETCH_STAT: (OBJECT_ID_NAME, OBJECT_NAME_NAME,
                                MODIFIED_DATE_NAME,
                                ORIGINAL_CONTENT_SIZE_NAME),
            }

        fetch_list = []
        for item in FETCH_ATTRS_DICT[fetch_type]:
            fetch_list.append((FETCH_KEY, item))
        return urlencode(fetch_list)

    def __get_object_id_dict(self, contents):
        """Return dictionary of the object id in "contents"."""

        object_id_dict = {}
        for content in contents.get(ITEMS_NAME, []):
            object_name = content.get(OBJECT_NAME_NAME, "")
            object_id = content[OBJECT_ID_NAME]
            if object_id.startswith(PRIVATE_DRAWER_ID_PREFIX):
                object_id_dict[PRIVATE_DRAWER_NAME] = object_id
                continue

            if object_name:
                object_id_dict[object_name] = object_id

        return object_id_dict

    def __get_child_path(self, current_path, full_path):
        """Return a child path of "current_path" under "full_path"."""

        child_path = full_path
        while current_path != os.path.dirname(child_path):
            child_path = os.path.dirname(child_path)

        return child_path

    def __init_object_ids(self):
        """Initialize each IDs."""

        self.drawer_id = ""
        self.path_id_map = PathIdMap()

    def __get_length(self, data):
        """Return a length of "data"."""

        try:
            return len(data)
        except:
            pass

        try:
            length = 0
            if hasattr(data, "seek") and hasattr(data, "tell"):
                data.seek(0, os.SEEK_END)
                length = data.tell()
                data.seek(0)

            return length
        except:
            pass

        return 0

    def __init__(self, base_url, cabinet_no, user, password, client_id,
            client_secret, drawer=PRIVATE_DRAWER_NAME):
        """Constructor."""

        self.base_url = base_url
        self.cabinet_no = cabinet_no
        self.user = user
        self.password = password
        self.client_id = client_id
        self.client_secret = client_secret
        self.drawer = drawer
        self.session_id = ""
        self.purpose = AccessInfo.Purpose.ACCESS_DIRECTORY
        self.access_base_url = "/".join(
                                        (self.base_url, self.cabinet_no))
        self.target_name = ""
        self.path_info_base_path = ""
        self.drawer_id = ""
        self.path_id_map = None
        self.__init_object_ids()

    def set_drawer(self, drawer=PRIVATE_DRAWER_NAME):
        """Set target drawer."""

        if self.drawer != drawer:
            self.drawer = drawer
            self.__init_object_ids()
        return self.drawer

    def set_session_info(self, response=None):
        """Set session information from data which is in "response"."""

        self.__init_object_ids()
        if response:
            self.session_id = self.__get_session_id(
                                response.info().getheader("Set-Cookie", ""))
        else:
            self.session_id = ""

    def set_path_info(self, response):
        """Set the path information which is in "response"."""

        contents = json.loads(self.__to_str(response))
        object_id_dict = self.__get_object_id_dict(contents)
        if not self.drawer_id:
            self.__init_object_ids()
            self.drawer_id = object_id_dict.get(self.drawer, "")
            if not self.drawer_id:
                raise DrawerNotFoundHttpError()
            self.path_id_map.append_path(ROOT_PATH, self.drawer_id)
            return

        for name in object_id_dict:
            self.path_id_map.append_path(
                                os.path.join(self.path_info_base_path, name),
                                object_id_dict[name])

        if self.target_name and not self.target_name in object_id_dict:
            raise ObjectNotFoundHttpError()

    def set_create_path_info(self, path, response):
        """Set the path information which is in "response" of request
        for creating the "path".
        """

        contents = json.loads(self.__to_str(response))
        object_id = contents.get(OBJECT_ID_NAME, "")
        if object_id:
            self.path_id_map.append_path(path, object_id)

    def to_drawer_list(self, response):
        """Convert "response" to the drawer list."""

        contents = json.loads(self.__to_str(response))
        content_list = []
        for content in contents.get(ITEMS_NAME, []):
            object_id = content[OBJECT_ID_NAME]
            if object_id.startswith(PRIVATE_DRAWER_ID_PREFIX):
                name = PRIVATE_DRAWER_NAME
            elif object_id.startswith(PUBLIC_DRAWER_ID_PREFIX):
                name = content[OBJECT_NAME_NAME]
            else:
                # Don't support other drawer
                continue

            description = ""
            if DESCRIPTION_NAME in content:
                description = content[DESCRIPTION_NAME]

            content_info = {NAME_NAME: name, DESCRIPTION_NAME: description}
            content_list.append(content_info)

        return content_list

    def to_content_list(self, response):
        """Convert "response" to the content list."""

        contents = json.loads(self.__to_str(response))
        content_list = []
        for content in contents.get(ITEMS_NAME, []):
            content_info = {NAME_NAME: content[OBJECT_NAME_NAME],
                            STAT_NAME: self.to_stat_info(json.dumps(content))}

            content_list.append(content_info)

        return content_list

    def convert_path(self, path):
        """Convert charcter in "path" according to the server rules."""

        is_decoded = False
        if isinstance(path, str):
            try:
                is_decoded = True
                path = path.decode("UTF-8")
            except:
                # Do nothing
                return path

        # Normalize "path" according to unicode "NFC" rules."""
        normalized_path = unicodedata.normalize("NFC", path)
        if is_decoded:
            normalized_path = normalized_path.encode("UTF-8")
        return normalized_path

    def compare_path(self, path, path_on_server):
        """Compare "path" and "path_on_server".
           "path" is converted according to the server rules.
        """

        return self.convert_path(path) == path_on_server

    def to_stat_info(self, response):
        """Convert "response" to the stat information."""

        json_stat_info = json.loads(self.__to_str(response))
        if json_stat_info[OBJECT_ID_NAME].startswith(FOLDER_ID_PREFIX):
            mode = DIR_MODE
            size = 0
        else:
            mode = FILE_MODE
            size = json_stat_info.get(ORIGINAL_CONTENT_SIZE_NAME, 0)

        utc_time_tuple = time.strptime(json_stat_info[MODIFIED_DATE_NAME],
                                        MTIME_FORMAT)
        mtime = int(calendar.timegm(utc_time_tuple))

        return {
                stat.ST_MODE: mode,
                stat.ST_MTIME: mtime,
                stat.ST_SIZE: size,
            }

    def get_open_session_req_info(self):
        """Return the "RequestInfo" to open session."""

        param = {
                "userId": self.user,
                "password": self.password,
                "clientId": self.client_id,
                "clientSecret": self.client_secret,
            }
        return RequestInfo(self.access_base_url + "/session/login",
                            json.dumps(param),
                            headers=self.__get_headers(DATA_JSON))

    def get_close_session_req_info(self):
        """Return the "RequestInfo" to close session."""

        return RequestInfo(self.access_base_url + "/session/logout",
                                    headers=self.__get_headers(DATA_JSON),
                            method=POST_METHOD)

    def get_prepare_path_req_info(self, path, purpose):
        """Return the "RequestInfo" to prepare accesing a "path"."""

        DIRECT_DIRECTORY_PURPOSE_LIST = [
                AccessInfo.Purpose.ACCESS_DIRECTORY,
                AccessInfo.Purpose.DELETE_DIRECTORY,
                AccessInfo.Purpose.RENAME_DIRECTORY,
                AccessInfo.Purpose.MOVE_DIRECTORY,
                AccessInfo.Purpose.MOVE_TO,
            ]

        DIRECT_FILE_PURPOSE_LIST = [
                AccessInfo.Purpose.ACCESS_FILE,
                AccessInfo.Purpose.DELETE_FILE,
                AccessInfo.Purpose.RENAME_FILE,
                AccessInfo.Purpose.MOVE_FILE,
            ]

        PARENT_DIRECTORY_PURPOSE_LIST = [
                AccessInfo.Purpose.MAKE_DIRECTORY,
                AccessInfo.Purpose.WRITE_FILE,
            ]

        self.purpose = purpose

        if not self.drawer_id:
            url = self.access_base_url + OBJECT_CHILDREN_URL_TMPL % \
                                        (CABINET_ID,)
            url += "?" + self.__get_fetch_param(FETCH_DRAWER)
            return RequestInfo(url, headers=self.__get_headers())

        target_path = ""
        if self.purpose in DIRECT_DIRECTORY_PURPOSE_LIST:
            target_path = path
            self.path_id_map.append_directory(target_path)
        elif self.purpose in DIRECT_FILE_PURPOSE_LIST:
            target_path = path
            self.path_id_map.append_path(target_path)
        elif self.purpose in PARENT_DIRECTORY_PURPOSE_LIST:
            target_path = os.path.dirname(path)
            self.path_id_map.append_directory(target_path)
        else:
            return None

        if self.path_id_map.get_id(target_path) != None:
            return None

        target_name = ""
        while True:
            target_name = os.path.basename(target_path)
            target_path = os.path.dirname(target_path)
            parent_id = self.path_id_map.get_id(target_path)
            if parent_id != None:
                break

        if target_name == ROOT_PATH:
            target_name = ""
        self.target_name = target_name
        self.path_info_base_path = target_path
        url = self.access_base_url + OBJECT_CHILDREN_URL_TMPL % (parent_id,)
        url += "?" + self.__get_fetch_param(FETCH_STAT)

        return RequestInfo(url, headers=self.__get_headers())

    def get_list_drawers_req_info(self):
        """Return the "RequestInfo" to list drawers."""

        url = self.access_base_url + OBJECT_CHILDREN_URL_TMPL % \
                                        (CABINET_ID,)
        url += "?" + self.__get_fetch_param(FETCH_DRAWER)
        return RequestInfo(url, headers=self.__get_headers())

    def get_list_contents_req_info(self, path):
        """Return the "RequestInfo" to list contents in "path" directory."""

        url = self.access_base_url + OBJECT_CHILDREN_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        url += "?" + self.__get_fetch_param(FETCH_STAT)
        return RequestInfo(url, headers=self.__get_headers())

    def get_make_directory_req_info(self, path):
        """Return the "RequestInfo" to make "path" directory."""

        url = self.access_base_url + MAKE_DIRECTORY_URL_TMPL % \
                            (self.path_id_map.get_id(os.path.dirname(path)),)
        param = {OBJECT_NAME_NAME: self.__to_utf8(os.path.basename(path))}
        return RequestInfo(url, json.dumps(param),
                            headers=self.__get_headers(DATA_JSON))

    def get_delete_directory_req_info(self, path):
        """Return the "RequestInfo" to delete "path" directory."""

        url = self.access_base_url + OBJECT_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        return RequestInfo(url, headers=self.__get_headers(),
                            method=DELETE_METHOD)

    def get_stat_directory_req_info(self, path):
        """Return the "RequestInfo" to get status of "path" directory."""

        url = self.access_base_url + OBJECT_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        url += "?" + self.__get_fetch_param(FETCH_STAT)
        return RequestInfo(url, headers=self.__get_headers())

    def get_rename_directory_req_info(self, path, new_name):
        """Return the "RequestInfo" to rename "path" directory."""

        url = self.access_base_url + OBJECT_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        param = {OBJECT_NAME_NAME: self.__to_utf8(new_name)}
        return RequestInfo(url, json.dumps(param),
                            headers=self.__get_headers(DATA_JSON),
                            method=PUT_METHOD)

    def get_move_directory_req_info(self, path, parent_path):
        """Return the "RequestInfo" to move "path" directory."""

        url = self.access_base_url + MOVE_OBJECT_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        param = {"destinationId": self.path_id_map.get_id(parent_path)}
        return RequestInfo(url, json.dumps(param),
                            headers=self.__get_headers(DATA_JSON))

    def get_delete_file_req_info(self, path):
        """Return the "RequestInfo" to delete "path" file."""

        url = self.access_base_url + OBJECT_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        return RequestInfo(url, headers=self.__get_headers(),
                            method=DELETE_METHOD)

    def get_stat_file_req_info(self, path):
        """Return the "RequestInfo" to get status of "path" file."""

        url = self.access_base_url + OBJECT_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        url += "?" + self.__get_fetch_param(FETCH_STAT)
        return RequestInfo(url, headers=self.__get_headers())

    def get_rename_file_req_info(self, path, new_name):
        """Return the "RequestInfo" to rename "path" file."""

        url = self.access_base_url + OBJECT_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        param = {OBJECT_NAME_NAME: self.__to_utf8(new_name)}
        return RequestInfo(url, json.dumps(param),
                            headers=self.__get_headers(DATA_JSON),
                            method=PUT_METHOD)

    def get_move_file_req_info(self, path, parent_path):
        """Return the "RequestInfo" to move "path" file."""

        url = self.access_base_url + MOVE_OBJECT_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        param = {"destinationId": self.path_id_map.get_id(parent_path)}
        return RequestInfo(url, json.dumps(param),
                            headers=self.__get_headers(DATA_JSON))

    def get_read_file_req_info(self, path):
        """Return the "RequestInfo" to read from "path" file."""

        url = self.access_base_url + FILE_READ_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        return RequestInfo(url, headers=self.__get_headers())

    def get_write_file_req_info(self, path, data, length=0, overwrite=False):
        """Return the "RequestInfo" to write to "path" file."""

        url = self.access_base_url + FILE_WRITE_URL_TMPL % \
                            (self.path_id_map.get_id(os.path.dirname(path)),)
        duplicationMode = "error"
        if overwrite:
            duplicationMode = "overwrite"
        param = {
                "documentName": self.__to_utf8(os.path.basename(path)),
                "duplicationMode": duplicationMode,
            }

        if length <= 0:
            length = self.__get_length(data)

        url += "?" + urlencode(param)
        return RequestInfo(
                url, data, headers=self.__get_headers(DATA_BIN, length))
